################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import pygame
import pyros
import pyros_support_ui.gccui
import threading
import time
import socket
import netifaces
import sys

from pyros_support_ui import gccui


DEBUG = False

rovers = []
do_discovery = True
show_rovers = False
selected_rover = 0
connected_rover = 0
selectedRoverMap = {}

# {
#     "rover2": {
#         "address": "172.24.1.184",
#         "port": 1883
#     },
#     "rover3": {
#         "address": "172.24.1.185",
#         "port": 1883
#     },
#     "rover4": {
#         "address": "172.24.1.186",
#         "port": 1883
#     },
#     "rover2proxy": {
#         "address": "gcc-wifi-ap",
#         "port": 1884
#     },
#     "rover3proxy": {
#         "address": "gcc-wifi-ap",
#         "port": 1885
#     },
#     "rover4proxy": {
#         "address": "gcc-wifi-ap",
#         "port": 1886
#     },
# }

THIS_PORT = 0xd15d
DISCOVERY_PORT = 0xd15c
BROADCAST_TIMEOUT = 5  # every 5 seconds
ROVER_TIMEOUT = 30  # 30 seconds no

sckt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
sckt.settimeout(1)


def setup_listen_socket():
    global THIS_PORT
    connected = 5

    while connected > 0:
        # noinspection PyBroadException
        try:
            sckt.bind(('', THIS_PORT))
            connected = 0
        except Exception:
            THIS_PORT = THIS_PORT + 1


setup_listen_socket()


def get_broadcasts():
    broadcasts = []

    ifacenames = netifaces.interfaces()
    for ifname in ifacenames:
        addrs = netifaces.ifaddresses(ifname)

        for d in addrs:
            for dx in addrs[d]:
                if "broadcast" in dx:
                    broadcasts.append(dx["broadcast"])
    return broadcasts


def add_to_list(rover_map):
    rover_map["lastSeen"] = time.time()
    if "PORT" in rover_map:
        rover_map["PORT"] = int(rover_map["PORT"])

    for rover in rovers:
        if rover["IP"] == rover_map["IP"] and rover["PORT"] == rover_map["PORT"]:
            for key in rover_map:
                rover[key] = rover_map[key]
            return

    rovers.append(rover_map)
    print("Found new rover " + str(rover_map["NAME"]) + " @ " + str(rover_map["IP"]) + ":" + str(rover_map["PORT"]))


def discover():
    global do_discovery
    packet = "Q#IP=255.255.255.255;PORT=" + str(THIS_PORT)

    last_broadcast_time = time.time()

    while True:
        if do_discovery or time.time() - last_broadcast_time > BROADCAST_TIMEOUT:
            do_discovery = False
            # noinspection PyBroadException
            try:
                broadcasts = get_broadcasts()

                for broadcast in broadcasts:
                    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                    bs = bytes(packet, 'utf-8')
                    s.sendto(bs, (broadcast, DISCOVERY_PORT))
                last_broadcast_time = time.time()

                for i in range(len(rovers) - 1, -1, -1):
                    rover = rovers[i]
                    if rover["lastSeen"] + ROVER_TIMEOUT < time.time():
                        print("Lost connection to rover " + str(rovers[i]["name"]) + " @ " + str(rovers[i]["ip"]) + ":" + str(rovers[i]["port"]))
                        del rovers[i]
            except Exception:
                pass

        # noinspection PyBroadException
        try:
            data, addr = sckt.recvfrom(1024)

            p = str(data, 'utf-8')

            if p.startswith("A#"):
                kvs = p[2:].split(";")

                ip = None
                name = None
                port = None
                device_type = None

                rover_map = {}

                for keyValue in kvs:
                    kvp = keyValue.split("=")
                    if len(kvp) == 2:
                        rover_map[kvp[0]] = kvp[1]
                        if kvp[0] == "IP":
                            ip = kvp[1]
                        elif kvp[0] == "PORT":
                            # noinspection PyBroadException
                            try:
                                port = int(kvp[1])
                            except Exception:
                                pass
                        elif kvp[0] == "NAME":
                            name = kvp[1]
                        elif kvp[0] == "TYPE":
                            device_type = kvp[1]
                        elif kvp[0] == "PYROS":
                            # noinspection PyBroadException
                            try:
                                port = int(kvp[1])
                            except Exception:
                                pass
                            device_type = "ROVER"

                if name is None:
                    name = ip
                if (port is not None) and (ip is not None) and device_type == "ROVER":
                    add_to_list(rover_map)
                    received_something = True

            print("Got " + p + "  Rovers: " + str(rovers))
        except Exception:
            pass


def add_supplied_rover():
    rover_map = {}

    kv = sys.argv[1].split(":")
    if len(kv) == 1:
        kv.append("1883")

    rover_map["IP"] = kv[0]
    rover_map["PORT"] = int(kv[1])
    rover_map["NAME"] = "Rover(args)"
    rover_map["TYPE"] = "ROVER"
    rover_map["JOY_PORT"] = "1880"

    add_to_list(rover_map)


if len(sys.argv) > 1:
    add_supplied_rover()

else:
    thread = threading.Thread(target=discover, args=())
    thread.daemon = True
    thread.start()


def get_host():
    if len(rovers) == 0:
        return None
    if DEBUG:
        print("Selected rover " + str(rovers[selected_rover]))
    return rovers[selected_rover]["IP"]


def get_port():
    if len(rovers) == 0:
        return None
    return int(rovers[selected_rover]["PORT"])


def connect():
    global selected_rover, connected_rover

    connected_rover = selected_rover

    if len(rovers) == 0:
        return None
    pyros.connect(rovers[selected_rover]["IP"], rovers[selected_rover]["PORT"], waitToConnect=False)


lmeta = False
rmeta = False
lshift = False
rshift = False


def handle_connect_key_down(key):
    global selected_rover, do_discovery, show_rovers
    global lmeta, rmeta, lshift, rshift

    if key == pygame.K_LMETA:
        lmeta = True
    elif key == pygame.K_RMETA:
        rmeta = True
    elif key == pygame.K_LSHIFT:
        lshift = True
    elif key == pygame.K_RSHIFT:
        rshift = True
    elif lmeta and key == pygame.K_q:
        sys.exit(0)

    if key == pygame.K_1 and (lmeta or rmeta):
        do_discovery = True
        show_rovers = True
    elif key == pygame.K_TAB:
        if show_rovers:
            if lshift or rshift:
                selected_rover -= 1
                if selected_rover < 0:
                    selected_rover = len(rovers) - 1
            else:
                selected_rover += 1
                if selected_rover >= len(rovers):
                    selected_rover = 0
        else:
            show_rovers = True
    elif show_rovers and key == pygame.K_UP and selected_rover > 0:
        selected_rover -= 1
    elif show_rovers and key == pygame.K_DOWN and selected_rover < len(rovers) - 1:
        selected_rover += 1
    elif show_rovers and key == pygame.K_RETURN or key == pygame.K_KP_ENTER:
        connect()
        show_rovers = False
    elif show_rovers and key == pygame.K_ESCAPE:
        show_rovers = False
    elif key == pygame.K_F5:
        connect()

    return show_rovers


def handle_connect_key_up(key):
    global lmeta, rmeta, lshift, rshift

    if key == pygame.K_LMETA:
        lmeta = False
    elif key == pygame.K_RMETA:
        rmeta = False
    elif key == pygame.K_LSHIFT:
        lshift = False
    elif key == pygame.K_RSHIFT:
        rshift = False

    return show_rovers


def get_selected_rover_details_text(i):
    if selected_rover >= len(rovers):
        return "No rovers"

    name = "Unknown Rover"

    if "NAME" in rovers[i]:
        name = rovers[i]["NAME"]

    if name.startswith("gcc-rover-"):
        name = "GCC Rover " + name[10:]

    return name


def get_selected_rover_ip(i):
    if selected_rover >= len(rovers):
        return ""

    return str(rovers[i]["IP"]) + ":" + str(rovers[i]["PORT"])


_connection_counter = 1
_state = 1


def _draw_red_indicator():
    if _state == 1:
        gccui.draw_rect((9, 10), (16, 16), (96, 0, 0, 128), 2, 0)
        gccui.draw_rect((11, 12), (11, 11), (160, 0, 0, 64), 2, 0)
        gccui.draw_filled_rect((13, 14), (8, 8), (255, 0, 0, 64))
    if _state == 0:
        gccui.draw_filled_rect((9, 10), (16, 16), (160, 0, 0, 128))
        gccui.draw_rect((11, 12), (11, 11), (180, 0, 0, 64), 2, 0)
        gccui.draw_filled_rect((13, 14), (8, 8), (255, 0, 0, 64))


def _draw_green_indicator():
    if _state == 0 or _state == 1:
        gccui.draw_filled_rect((9, 10), (16, 16), (0, 128, 0, 128))
        gccui.draw_rect((11, 12), (11, 11), (0, 180, 0, 64), 2, 0)
        gccui.draw_filled_rect((13, 14), (8, 8), (0, 255, 0, 64))
    if _state == 2:
        gccui.draw_filled_rect((11, 10), (16, 16), (0, 96, 0, 128))
        gccui.draw_rect((9, 12), (15, 11), (0, 160, 0, 64), 2, 0)
        gccui.draw_filled_rect((11, 14), (12, 8), (0, 220, 0, 64))


def draw_connection():

    def draw_rover(rover_index, pos, colour):
        w = gccui.font.size(get_selected_rover_details_text(rover_index))[0]
        gccui.screen.blit(gccui.font.render(get_selected_rover_details_text(rover_index), 1, colour), pos)
        gccui.screen.blit(gccui.smallFont.render(get_selected_rover_ip(rover_index), 1, colour), (pos[0] + 8 + w, pos[1] + 5))

    global _connection_counter, _state
    if pyros.isConnected():
        _connectionCounter -= 1
        if _connectionCounter < 0:
            if _state == 0:
                _state = 2
                _connectionCounter = 10
            elif _state == 1:
                _state = 0
                _connectionCounter = 10
            elif _state == 2:
                _state = 1
                _connectionCounter = 10

        _draw_green_indicator()
        draw_rover(connected_rover, (32, 0), gccui.GREEN)
    else:
        _connectionCounter -= 1
        if _connectionCounter < 0:
            _connectionCounter = 10
            _state -= 1
            if _state < 0:
                _state = 3

        _draw_red_indicator()
        draw_rover(connected_rover, (32, 0), gccui.RED)

    if show_rovers:
        gccui.draw_frame(pygame.Rect(0, 30, 250, 178), gccui.LIGHT_BLUE, gccui.BLACK)

        st = 0
        if selected_rover > 2:
            st = selected_rover - 2

        en = st + 5
        if en > len(rovers):
            en = len(rovers)

        while en - st < 5 and st > 0:
            st = st - 1

        for i in range(st, en):
            y = 30 + (i - st) * 30 + 12
            if i == selected_rover:
                pygame.draw.rect(gccui.screen, gccui.DARK_GREY, pygame.Rect(8, y, 234, 30), 0)

            draw_rover(i, (12, y), gccui.LIGHT_BLUE)

        if st > 0:
            gccui.draw_up_arrow(120, 38, 130, 42, gccui.LIGHT_BLUE)
        if en < len(rovers):
            gccui.draw_down_arrow(120, 194, 130, 198, gccui.LIGHT_BLUE)



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import pygame
import pyros
import threading
import traceback
import time

from functools import partial

from pygame import Rect, draw

from pyros_support_ui.components import Component, Collection, Button
from discovery import Discover


_GREEN = (128, 255, 128)
_RED = (255, 128, 128)
_YELLOW = (255, 255, 128)
_BLUE = (128, 128, 255)
_LIGHT_BLUE = (160, 255, 255)
_BLACK = (0, 0, 0)
_DARK_GREY = (80, 80, 80)
_GREY = (160, 160, 160)
_WHITE = (255, 255, 255)


class _DiscoveredRover:
    def __init__(self, host: str, port: int, name: str):
        self.host = host
        self.port = port
        self.name = name


class ScreenFrame(Component):
    def __init__(self, rect,
                 line_colour, background_colour,
                 margin=4,
                 background_image=None,
                 logo_image=None, logo_alt_image=None):
        super(ScreenFrame, self).__init__(rect)
        self.background_image = background_image
        self.logo_image = logo_image
        self.logo_alt_image = logo_alt_image if logo_alt_image else logo_image
        self.line_colour = line_colour
        self.background_colour = background_colour
        self.margin = margin

    def draw(self, surface):
        if self.background_image is not None:
            surface.blit(self.background_image, (0, 0))
        else:
            if self.background_colour is not None:
                draw.rect(surface, self.background_colour, self.rect, 0)
            else:
                surface.fill((0, 0, 0))

        rect = self.rect
        draw.rect(surface, self.line_colour, (rect.x + self.margin, rect.y + self.margin, rect.width - self.margin * 2, rect.height - self.margin * 2), 1)
        for _i in range(6, 12):
            draw.line(surface, self.line_colour, (rect.x + rect.width - _i, rect[1] + rect.height - self.margin - 2), (rect.x + rect.width - self.margin - 2, rect.y + rect.height - _i), 1)

        if self.logo_image is not None:
            image_size = self.logo_image.get_size()
            if int(time.time()) % 20 < 10:
                surface.blit(self.logo_image, (rect.x + rect.width - self.margin * 2 - self.logo_image.get_width(), rect.y + self.margin * 2))
            else:
                surface.blit(self.logo_alt_image, (rect.x + rect.width - self.margin * 2 - self.logo_alt_image.get_width(), rect.y + self.margin * 2))
        else:
            image_size = (32, 32)

        draw.line(surface, self.line_colour, (rect.x + rect.width - image_size[0] - self.margin, rect.y + image_size[1] + 12),
                  (rect.x + rect.width - self.margin, rect.y + image_size[1] + 12))
        draw.line(surface, self.line_colour, (rect.x + rect.width - image_size[0] - self.margin, rect.y + image_size[1] + 12),
                  (rect.x + rect.width - image_size[0] - 12, rect.y + image_size[1] + self.margin))

        draw.line(surface, self.line_colour, (rect.x + rect.width - image_size[0] - 12, rect.y + self.margin),
                  (rect.x + rect.width - image_size[0] - 12, rect.y + image_size[1] + self.margin))


class PyrosServersButton(Button):
    def __init__(self, rect, colour,
                 slate=30,
                 font=None,
                 small_font=None,
                 on_click=None):

        super(PyrosServersButton, self).__init__(rect, on_click=on_click)  # Call super constructor to store rectangle
        self.colour = colour
        self.font = font
        self.small_font = small_font
        self.slate = slate
        self._state = 1
        self._connection_counter = 1
        self.rovers = []
        self.selected_rover = None

    def draw(self, surface):
        def _draw_rect(pos, size, colour, stick=0, outside=1):
            draw.line(surface, colour, (pos[0] - stick, pos[1] - outside), (pos[0] + size[0] + stick, pos[1] - outside))
            draw.line(surface, colour, (pos[0] - stick, pos[1] + size[1] + outside), (pos[0] + size[0] + stick, pos[1] + size[1] + outside))

            draw.line(surface, colour, (pos[0] - outside, pos[1] - stick), (pos[0] - outside, pos[1] + size[1] + stick))
            draw.line(surface, colour, (pos[0] + size[0] + outside, pos[1] - stick), (pos[0] + size[0] + outside, pos[1] + size[1] + stick))

        def _draw_filled_rect(pos, size, colour=_LIGHT_BLUE):
            draw.rect(surface, colour, pygame.Rect(pos, size))

        def _draw_red_indicator():
            if self._state == 1:
                _draw_rect((9, 10), (16, 16), (96, 0, 0, 128), 2, 0)
                _draw_rect((11, 12), (11, 11), (160, 0, 0, 64), 2, 0)
                _draw_filled_rect((13, 14), (8, 8), (255, 0, 0, 64))
            if self._state == 0:
                _draw_filled_rect((9, 10), (16, 16), (160, 0, 0, 128))
                _draw_rect((11, 12), (11, 11), (180, 0, 0, 64), 2, 0)
                _draw_filled_rect((13, 14), (8, 8), (255, 0, 0, 64))

        def _draw_green_indicator():
            if self._state == 0 or self._state == 1:
                _draw_filled_rect((9, 10), (16, 16), (0, 128, 0, 128))
                _draw_rect((11, 12), (11, 11), (0, 180, 0, 64), 2, 0)
                _draw_filled_rect((13, 14), (8, 8), (0, 255, 0, 64))
            if self._state == 2:
                _draw_filled_rect((11, 10), (16, 16), (0, 96, 0, 128))
                _draw_rect((9, 12), (15, 11), (0, 160, 0, 64), 2, 0)
                _draw_filled_rect((11, 14), (12, 8), (0, 220, 0, 64))

        def _get_selected_rover_ip(rover):
            return rover.host + ":" + str(rover.port) if rover is not None else ""

        def _draw_rover(rover, pos, colour):
            rover_details_text = rover.name if rover is not None else ("No Rovers" if len(self.rovers) == 0 else "Select Rover")
            if len(rover_details_text) > 0:
                w = self.font.size(rover_details_text)[0]
                sur = self.font.render(rover_details_text, 1, colour)
                surface.blit(sur, pos)
                surface.blit(self.small_font.render(_get_selected_rover_ip(rover), 1, colour), (pos[0] + 8 + w, pos[1] + 5))

        rect = self.rect

        draw.line(surface, self.colour, (rect.x, rect.y), (rect.x + rect.width, rect.y))
        draw.line(surface, self.colour, (rect.x, rect.y), (rect.x, rect.y + rect.height))
        draw.line(surface, self.colour, (rect.x, rect.y + self.slate), (rect.x + rect.width - self.slate, rect.y + self.slate))
        draw.line(surface, self.colour, (rect.x + rect.width - self.slate, rect.y + self.slate), (rect.x + rect.width, rect.y))

        if pyros.is_connected():
            self._connection_counter -= 1
            if self._connection_counter < 0:
                if self._state == 0:
                    self._state = 2
                    _connection_counter = 10
                elif self._state == 1:
                    self._state = 0
                    self._connection_counter = 10
                elif self._state == 2:
                    self._state = 1
                    self._connection_counter = 10
                else:
                    self._state = 0

            _draw_green_indicator()
            _draw_rover(self.selected_rover, (32, 0), _GREEN)
        else:
            self._connection_counter -= 1
            if self._connection_counter < 0:
                self._connection_counter = 10
                self._state -= 1
                if self._state < 0:
                    self._state = 3

            _draw_red_indicator()
            _draw_rover(self.selected_rover, (32, 0), _RED)


class PyrosClientApp(Collection):
    _DISCOVERY_TIMEOUT = 30  # Every 10 seconds

    def __init__(self, ui_factory,
                 background_image=None,
                 logo_image=None, logo_alt_image=None,
                 font=None,
                 small_font=None,
                 connect_to_first=False,
                 connect_to_only=True):
        super(PyrosClientApp, self).__init__(None)  # Call super constructor to store rectable
        self.ui_factory = ui_factory
        self.ui_adapter = ui_factory.get_ui_adapter()
        self.content = None
        self.background_image = background_image
        self.logo_image = logo_image
        self.logo_alt_image = logo_alt_image if logo_alt_image else logo_image
        self.line_colour = ui_factory.colour
        self.font = font if font is not None else self.ui_factory.get_font()
        self.small_font = small_font if small_font is not None else self.ui_factory.get_small_font()
        
        self.screen_frame = ScreenFrame(None, self.line_colour, ui_factory.background_colour,
                                        background_image=background_image,
                                        logo_image=logo_image, logo_alt_image=logo_alt_image)
        self.add_component(self.screen_frame)

        self.discovery = Discover(5)
        self.thread = None
        self.connect_to_first = connect_to_first
        self.connect_to_only = connect_to_only
        self._first_discovery_run = True
        self.rovers_button = PyrosServersButton(Rect(0, 0, 100, 30), ui_factory.colour,
                                                font=self.font,
                                                small_font=self.small_font,
                                                on_click=self._show_rovers_action)
        self.add_component(self.rovers_button)

        self.rovers_menu = ui_factory.menu(Rect(0, 0, 150, 200), background_colour=pygame.color.THECOLORS['black'])
        # self.rovers_menu.add_menu_item("item1", partial(self._select_rover, 0), height=30)
        # self.rovers_menu.add_menu_item("item2", partial(self._select_rover, 1), height=30)
        self.add_component(self.rovers_menu)
        self.on_connected_subscribers = []
        self.redefine_rect(self.ui_adapter.get_screen().get_rect())

    def _show_rovers_action(self, _button, _pos):
        self.rovers_menu.show()

    def _select_rover(self, rover, _button, _pos):
        print(f"Selected: rover {rover.host}:{rover.port} {rover.name}")
        if self.rovers_button.selected_rover != rover:
            print(f"  Connecting rover {rover.host}:{rover.port} {rover.name}")
            pyros.connect(rover.host, rover.port, wait_to_connect=False)
            print(f"  Connected rover {rover.host}:{rover.port} {rover.name}")
            self.rovers_button.selected_rover = rover
        elif not pyros.is_connected():
            print(f"  Reconnecting rover {rover.host}:{rover.port} {rover.name}")
            pyros.connect(rover.host, rover.port, wait_to_connect=False)
            print(f"  Connected rover {rover.host}:{rover.port} {rover.name}")
        else:
            print(f"  Selected rover {rover.host}:{rover.port} {rover.name} - is already selected")
        self.rovers_menu.hide()

    def _set_contents_rectangle(self, rect):
        inside = Rect(rect.x + self.screen_frame.margin + 2,
                      self.rovers_button.rect.bottom + self.screen_frame.margin,
                      rect.width - self.screen_frame.margin * 2 - 4,
                      rect.bottom - self.rovers_button.rect.bottom - self.screen_frame.margin * 2 - 2)
        self.content.redefine_rect(inside)

    def redefine_rect(self, rect):
        # super(RoverClientApp, self).redefine_rect(rect)
        self.rect = rect
        self.screen_frame.redefine_rect(rect)
        self.rovers_button.redefine_rect(Rect(rect.x + self.screen_frame.margin, rect.y + self.screen_frame.margin, 320, 30))
        self.rovers_menu.redefine_rect(Rect(self.rovers_button.rect.x, self.rovers_button.rect.bottom, self.rovers_button.rect.width - self.rovers_button.slate * 2, 10))
        if self.content is not None:
            self._set_contents_rectangle(rect)

    def set_content(self, component):
        self.content = component
        top = self.components[-1]
        del self.components[-1]
        self.add_component(component)
        self.add_component(top)
        self._set_contents_rectangle(self.rect)

    def draw_border(self, surface):
        if self.background_image is not None:
            surface.blit(self.background_image, (0, 0))
        else:
            surface.fill((0, 0, 0))

        screen_rect = self.rect

        if self.logo_image is not None:
            image_size = self.logo_image.get_size()
            if int(time.time()) % 20 < 10:
                surface.blit(self.logo_image, (screen_rect.x + screen_rect.width - 8 - self.logo_image.get_width(), screen_rect.y + 8))
            else:
                surface.blit(self.logo_alt_image, (screen_rect.x + screen_rect.width - 8 - self.logo_alt_image.get_width(), screen_rect.y + 8))
        else:
            image_size = (32, 32)

        draw.line(surface, self.line_colour, (screen_rect.x + 4, screen_rect.y + 4), (screen_rect.x + 330, screen_rect.y + 4))
        draw.line(surface, self.line_colour, (screen_rect.x + 4, screen_rect.y + 4), (screen_rect.x + 4, screen_rect.y + 30))
        draw.line(surface, self.line_colour, (screen_rect.x + 4, screen_rect.y + 30), (screen_rect.x + 300, screen_rect.y + 30))
        draw.line(surface, self.line_colour, (screen_rect.x + 300, screen_rect.y + 30), (screen_rect.x + 330, screen_rect.y + 4))

        draw.line(surface, self.line_colour, (screen_rect.x + 330, screen_rect.y + 4), (screen_rect.x + screen_rect.width - 12, screen_rect.y + 4))

        draw.line(surface, self.line_colour, (screen_rect.x + screen_rect.width - 12, screen_rect.y + 4), (screen_rect.x + screen_rect.width - 4, screen_rect.y + 12))

        draw.line(surface, self.line_colour, (screen_rect.x + screen_rect.width - 4, screen_rect.y + 12), (screen_rect.x + screen_rect.width - 4, screen_rect.y + screen_rect.height - 4))
        draw.line(surface, self.line_colour, (screen_rect.x + 4, screen_rect.y + screen_rect.height - 4), (screen_rect.x + screen_rect.width - 4, screen_rect.y + screen_rect.height - 4))
        draw.line(surface, self.line_colour, (screen_rect.x + 4, screen_rect.y + screen_rect.height - 4), (screen_rect.x + 4, screen_rect.y + 30))

        draw.line(surface, self.line_colour, (screen_rect.x + screen_rect.width - image_size[0] - 4, screen_rect.y + image_size[1] + 12),
                  (screen_rect.x + screen_rect.width - 4, screen_rect.y + image_size[1] + 12))
        draw.line(surface, self.line_colour, (screen_rect.x + screen_rect.width - image_size[0] - 4, screen_rect.y + image_size[1] + 12),
                  (screen_rect.x + screen_rect.width - image_size[0] - 12, screen_rect.y + image_size[1] + 4))

        draw.line(surface, self.line_colour, (screen_rect.x + screen_rect.width - image_size[0] - 12, screen_rect.y + 4),
                  (screen_rect.x + screen_rect.width - image_size[0] - 12, screen_rect.y + image_size[1] + 4))

    def draw(self, surface):
        super(PyrosClientApp, self).draw(surface)
        # self.draw_border(surface)
        # self.content.draw(surface)
        # self.draw_top(surface)

    def pyros_init(self, pyros_mqtt_client_name, unique=True):
        pyros.init(pyros_mqtt_client_name, unique=unique, host=None, port=1883, wait_to_connect=False, on_connected=self._pyros_connected)

    def _pyros_connected(self):
        for on_connected_subscriber in self.on_connected_subscribers:
            try:
                on_connected_subscriber()
            except Exception as ex:
                print("MainLoop Exception: " + str(ex) + "\n" + ''.join(traceback.format_tb(ex.__traceback__)))

    def add_on_connected_subscriber(self, subscriber):
        self.on_connected_subscribers.append(subscriber)

    def remove_on_connected_subscriber(self, subscriber):
        self.on_connected_subscribers.remove(subscriber)

    def run_discovery(self):
        def process_name(_host, _port, name):
            return "Unknown Rover" if name is None else name.replace("gcc-rover-", "GCC Rover ")

        def make_rover(host, port, name):
            # noinspection PyBroadException
            try:
                _port = int(port)
                return _DiscoveredRover(host, _port, process_name(host, port, name))
            except Exception:
                return None

        def respomse_to_rover(response):
            rover = None
            if 'IP' in response and 'PYROS' in response:
                rover = make_rover(response['IP'], response['PYROS'], process_name(response['IP'], response['PYROS'], response.get('NAME')))
            elif 'IP' in response and 'PORT' in response and 'TYPE' in response and response['TYPE'] == 'ROVER':
                # backward compatibility
                rover = make_rover(response['IP'], response['PORT'], process_name(response['IP'], response['PORT'], response.get('NAME')))
            # else:
            #     rover = make_rover(response['IP'], 1883, process_name(response['IP'], 1883, response.get('NAME')))
            return rover

        def connect_to_first_callback(response):
            rover = respomse_to_rover(response)
            self.rovers_button.selected_rover = rover
            pyros.connect(rover.host, rover.port, wait_to_connect=False)

        callback = None
        if self._first_discovery_run and self.connect_to_first:
            callback = connect_to_first_callback
        responses = self.discovery.discover(callback)

        existing_rover_ids = set([r.host + str(r.port) + (r.name if r.name is not None else "") for r in self.rovers_button.rovers])
        new_rovers = []
        for response in responses:
            rover = respomse_to_rover(response)
            if rover is not None:
                new_rovers.append(rover)

        new_rover_ids = set([r.host + str(r.port) + (r.name if r.name is not None else "") for r in new_rovers])

        if existing_rover_ids != new_rover_ids:
            self.rovers_menu.clear_menu_items()
            del self.rovers_button.rovers[:]
            for rover in new_rovers:
                self.rovers_button.rovers.append(rover)
                self.rovers_menu.add_menu_item(rover.name, partial(self._select_rover, rover), height=30)

        if self._first_discovery_run:
            self._first_discovery_run = False
            if self.connect_to_only and not self.connect_to_first and len(self.rovers_button.rovers) == 1:
                rover = self.rovers_button.rovers[0]
                self.rovers_button.selected_rover = rover
                pyros.connect(rover.host, rover.port, wait_to_connect=False)

    def start_discovery_in_background(self):
        def _run_discovery_thread():
            last_run = 0
            while True:
                # noinspection PyBroadException
                try:
                    now = time.time()
                    if now - last_run < self._DISCOVERY_TIMEOUT:
                        time.sleep(now - last_run)
                    last_run = now
                    self.run_discovery()
                except Exception as ex:
                    pass  # Ignore errors
                    print("MainLoop Exception: " + str(ex) + "\n" + ''.join(traceback.format_tb(ex.__traceback__)))

        self.thread = threading.Thread(target=_run_discovery_thread, daemon=True)
        self.thread.start()

    def key_down(self, code):
        if code == pygame.K_TAB:
            if self.rovers_menu.is_visible():
                self.rovers_menu.hide()
            else:
                self._show_rovers_action(None, None)
            return True

        if self.rovers_menu.is_visible():
            return self.rovers_menu.key_down(code)
        return False

    def key_up(self, code):
        if self.rovers_menu.is_visible():
            return self.rovers_menu.key_up(code)
        return False



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import pygame
from local_resources import Resource
from io import BytesIO
from io import StringIO

last_keys = []
keys = []


def process_keys(on_key_down, on_key_up):
    global last_keys, keys

    last_keys = keys
    keys = pygame.key.get_pressed()
    if not len(keys) == 0 and not len(last_keys) == 0:

        for i in range(0, len(keys)):

            if keys[i] and not last_keys[i]:
                on_key_down(i)
            if not keys[i] and last_keys[i]:
                on_key_up(i)


def load_font(font_name: str, font_size: int):
    with Resource(font_name) as f:
        return pygame.font.Font(BytesIO(f.read()), font_size)


def load_image(image_name: str):
    with Resource(image_name) as f:
        return pygame.image.load(BytesIO(f.read()))





################################################################################
# Copyright (C) 2016-2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="pyros-support-ui",
    version="1.0.0",
    author="Daniel Sendula",
    description="Support library for (pygame based) UI components",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Abstract-Horizon/pyros-support-ui",
    packages=['pyros_support_ui'],
    zip_safe=False, # Doesn't create an egg - easier to debug and hack on
    install_requires=['pygame', 'pyros_client','local_resources', 'pyros-discovery'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################


from pyros_support_ui.components import *


class BackgroundChangeDecoration(Component):
    def __init__(self, colour):
        super(BackgroundChangeDecoration, self).__init__(None)  # Call super constructor to store rectable
        self.colour = colour

    def draw(self, surface):
        pygame.draw.rect(surface, self.colour, self.rect)


class MenuButtonBackgroundDecoration(Component):
    def __init__(self, colour):
        super(MenuButtonBackgroundDecoration, self).__init__(None)  # Call super constructor to store rectangle
        self.colour = colour

    def draw(self, surface):
        pygame.draw.rect(surface, self.colour, self.rect)


class ButtonRectangleDecoration(Component):
    def __init__(self, colour, background_colour):
        super(ButtonRectangleDecoration, self).__init__(None)  # Call super constructor to store rectangle
        self.colour = colour
        self.background_colour = background_colour
        self.strip_width = 5
        self.strip_margin = 3
        self.cut_width = 6
        self.cut_margin = 3
        self.cut_size = 15

    def draw(self, surface):
        if self.background_colour is not None:
            pygame.draw.rect(surface, self.background_colour, self.rect)

        x1 = self.rect.x
        x2 = self.rect.right - 1
        y1 = self.rect.y
        y2 = self.rect.bottom - 1
        try:
            pygame.draw.rect(surface, self.colour, pygame.Rect(x1, y1, self.strip_width, self.rect.height))
        except Exception:
            print(self.colour)
        pygame.draw.line(surface, self.colour, (x1 + self.strip_width + self.strip_margin, y1), (x2 - self.cut_size, y1))
        pygame.draw.line(surface, self.colour, (x2 - self.cut_size, y1), (x2, y1 + self.cut_size))
        pygame.draw.line(surface, self.colour, (x2, y1 + self.cut_size), (x2, y2))
        pygame.draw.line(surface, self.colour, (x2, y2), (x1 + self.strip_margin, y2))
        pygame.draw.line(surface, self.colour, (x1 + self.strip_width + self.strip_margin, y2), (x1 + self.strip_width + self.strip_margin, y1))

        pygame.draw.polygon(surface, self.colour, [
            (x2 - self.cut_size + self.cut_margin, y1),
            (x2 - self.cut_size + self.cut_margin + self.cut_width, y1),
            (x2, y1 + self.cut_size - self.cut_margin - self.cut_width),
            (x2, y1 + self.cut_size - self.cut_margin)
        ])


class BorderDecoration(Component):
    def __init__(self, rect, colour):
        super(BorderDecoration, self).__init__(rect)  # Call super constructor to store rectangle
        self.colour = colour
        self.top_left_cut = 8
        self.bottom_right_cut = 5

    def draw(self, surface):
        x1 = self.rect.x
        x2 = self.rect.right - 1
        y1 = self.rect.y
        y2 = self.rect.bottom - 1

        pygame.draw.polygon(surface, self.colour, [
            (x1 + self.top_left_cut, y1),
            (x2, y1),
            (x2, y2 - self.bottom_right_cut),
            (x2 - self.bottom_right_cut, y2),
            (x1, y2),
            (x1, y1 + self.top_left_cut)
        ], 1)


class BoxBlueSFThemeFactory(BaseUIFactory):
    def __init__(self, ui_adapter,
                 font=None,
                 small_font=None,
                 colour=pygame.color.THECOLORS['cornflowerblue'],
                 background_colour=(0, 0, 0, 255),
                 mouse_over_colour=pygame.color.THECOLORS['yellow'],
                 mouse_over_background_colour=pygame.color.THECOLORS['gray32']):
        super(BoxBlueSFThemeFactory, self).__init__(ui_adapter,
                                                    font=font,
                                                    small_font=small_font,
                                                    colour=colour,
                                                    background_colour=background_colour,
                                                    mouse_over_colour=mouse_over_colour)
        self.mouse_over_background_colour = mouse_over_background_colour

    def setMouseOverColour(self, mouse_over_colour):
        self.mouse_over_colour = mouse_over_colour

    def setBackgroundColour(self, background_colour):
        self.background_colour = background_colour

    def label(self, rect, text, font=None, colour=None, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL):
        label = Label(rect, text, font=font if font is not None else self.font, colour=colour, h_alignment=h_alignment, v_alignment=v_alignment)
        return label

    def image(self, rect, image, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL):
        return Image(rect, image, h_alignment=h_alignment, v_alignment=v_alignment)

    def button(self, rect, on_click=None, on_hover=None, label=None, hint=UiHint.NORMAL):
        colour = self.colour
        background_colour = self.background_colour
        mouse_over_colour = self.mouse_over_colour
        mouse_over_background_colour = self.mouse_over_background_colour
        if hint == UiHint.WARNING:
            colour = pygame.color.THECOLORS['orange']
            background_colour = self.background_colour
            mouse_over_colour = pygame.color.THECOLORS['orange']
            mouse_over_background_colour = pygame.color.THECOLORS['darkorange4']
        elif hint == UiHint.ERROR:
            colour = pygame.color.THECOLORS['red']
            background_colour = self.background_colour
            mouse_over_colour = pygame.color.THECOLORS['red']
            mouse_over_background_colour = pygame.color.THECOLORS['darkred']
        if hint == UiHint.NO_DECORATION:
            return Button(rect, on_click, on_hover, label,
                          mouse_over_decoration=BackgroundChangeDecoration(mouse_over_background_colour))
        else:
            return Button(rect, on_click, on_hover, label,
                          background_decoration=ButtonRectangleDecoration(colour, background_colour),
                          mouse_over_decoration=ButtonRectangleDecoration(mouse_over_colour, mouse_over_background_colour))

    def panel(self, rect, background_colour=None, hint=UiHint.NORMAL):
        if background_colour is None:
            if hint == UiHint.WARNING:
                background_colour = pygame.color.THECOLORS['orange']
            elif hint == UiHint.ERROR:
                background_colour = pygame.color.THECOLORS['red']
        return Panel(rect, background_colour, decoration=BorderDecoration(rect, self.colour))

    def menu(self, rect, background_colour=None, hint=UiHint.NORMAL):
        if background_colour is None:
            if hint == UiHint.WARNING:
                background_colour = pygame.color.THECOLORS['orange']
            elif hint == UiHint.ERROR:
                background_colour = pygame.color.THECOLORS['red']
        return Menu(rect, self, background_colour, decoration=BorderDecoration(rect, self.colour))

    def _menu_item_text_button(self, rect, label, callback):
        return self._menu_item_button(rect, self.label(None, label, h_alignment=ALIGNMENT.CENTER, v_alignment=ALIGNMENT.MIDDLE), callback)

    def _menu_item_button(self, rect, label, callback):
        return Button(rect, callback, label=label, mouse_over_decoration=MenuButtonBackgroundDecoration(self.mouse_over_background_colour))

    def border(self, rect, colour=pygame.color.THECOLORS['white']):
        return BorderDecoration(rect, colour)
